#! /usr/bin/python

"""
The addressbook holds a list of permission slips for recipients to which we
are allowed to send mail, and incoming permissions for senders from whom we
will accept mail.

Each OwnerEntry has a separate AddressBook.
"""

from petmail import util, permission, transport, crypto, pbutil, request
from petmail.interfaces import IAddressBook, IIDRecord, IAddressBookEntry, \
     IOwnerEntry, IOwnerPermission, IPermissionSlip
from twisted.python.components import implements
from twisted.python import log
from twisted.internet import defer

import os

class AddressBook(object):
    # TODO: needs test coverage
    __implements__ = IAddressBook
    debug = 0

    def __init__(self, owner):
        assert implements(owner, IOwnerEntry)
        self.owner = owner
        self.entries = []
        self.watchers = []

    def __getstate__(self):
        d = self.__dict__.copy()
        d['watchers'] = []
        return d

    def notifyOnChange(self, callback):
        """Register a callback to be run when the addressbook changes (such
        as when an entry is added, perhaps due to receipt of a message from
        a new sender). The callback is run with one argument, this
        AddressBook instance. The callback is not persisted."""
        self.watchers.append(callback)
    def dontNotifyOnChange(self, callback):
        self.watchers.remove(callback)
    def notifyWatchers(self):
        for w in self.watchers:
            w(self)

    def lookupID(self, ID):
        if not isinstance(ID, type("")):
            print "AddressBook.lookupID got %s (type %s), not string" % \
                  (ID, type(ID))
            assert False
        for e in self.entries:
            if e.idRecord and e.idRecord.ID == ID:
                return e
        return None
    def lookupPetName(self, name):
        for e in self.entries:
            if e.petname == name:
                return e
        return None
    def lookupAddress(self, address):
        for e in self.entries:
            if address in e.idRecord.addresses:
                return e

    def addEntry(self, entry, message=None):
        # message is the message which caused this entry to be added
        # TODO: stash (parts of) the message
        assert(isinstance(entry, AddressBookEntry))
        entry.addressbook = self
        self.entries.append(entry)
        self.notifyWatchers()

    def addRecord(self, record):
        # this method is (currently) only used by the gui in response to an
        # explicit address lookup, using Agent.retrieveAddress()
        assert implements(record, IIDRecord)
        e = self.lookupID(record.ID)
        if e:
            # TODO: a bit silly, as the lookup method already did a
            # verify/decode of the IDRecord.
            e.idRecord.updateFromSigned(record.signed)
            if self.debug:
                print "addRecord", record.genericPermission
        else:
            e = AddressBookEntry()
            e.idRecord = record
            self.entries.append(e)
        self.notifyWatchers()
        return e

    def deleteEntry(self, entry):
        if entry.theirPermission:
            entry.revokeTheirPermission()
            # would normally do entry.modified() here, but we need to push
            # out the update immediately instead of waiting for the holdoff
            # time
            self.owner.parent.refreshRecipient(self.owner.ID,
                                               entry.idRecord.ID)
        self.entries.remove(entry)
        self.notifyWatchers()

class AddressBookEntry(object):
    __implements__ = IAddressBookEntry
    updateHoldoffTime = 5*60
    doUpdate = True

    def __init__(self):
        self.addressbook = None # set by AddressBook.addEntry
        self.petname = None
        self.idRecord = None
        self.ourPermission = None   # what we use to send mail to them
        # .ourPermission is always a PermissionSlip
        self.theirPermission = None # what we require when they send to us
        # .theirPermission is always an OwnerPermission
        self.doNotGrantReturnPermission = False
        self.lastSentOwnerEntry = 0 # the last time we sent them our IDRecord
        self.lastSentTheirPermission = 0 # last time we sent .theirPermission
        self.watchers = []

    def needToSendTheirPermission(self, debug=False):
        p = self.theirPermission
        if not p:
            if self.lastSentTheirPermission:
                # revoked
                if debug: print "permission was revoked, must update"
                return True
            if debug: print "no permission to update"
        else:
            if debug:
                print "last sent %d, lastModified %d" % \
                      (self.lastSentTheirPermission, p.lastModified)
            if self.lastSentTheirPermission < p.lastModified:
                if debug: print "permission changed recently, must update"
                return True
            if debug: print "permission is up-to-date"
        return False
    def sentTheirPermission(self, when):
        if self.theirPermission:
            self.lastSentTheirPermission = when
        else:
            self.lastSentTheirPermission = 0

    def __getstate__(self):
        d = self.__dict__.copy()
        d['watchers'] = []
        return d

    def notifyOnChange(self, callback):
        """Register a callback to be run when the entry changes (such as
        when the permission is modified, perhaps due to receipt of a message
        from the recipient sender). The callback is run with one argument,
        this AddressBookEntry instance. The callback is not persisted."""
        self.watchers.append(callback)
    def dontNotifyOnChange(self, callback):
        self.watchers.remove(callback)
    def notifyWatchers(self):
        for w in self.watchers:
            w(self)

    def setPetName(self, name):
        self.petname = name

    def createIDRecord(self, signed, cagent, cause):
        """Create our .idRecord with the signed IDRecord they just sent us.
        Return a Deferred that will fire (with this entry as an argument)
        when the record is complete."""
        # TODO: record 'cause' as the cause of the update
        assert not self.idRecord
        self.idRecord = IDRecord()
        d = self.idRecord.addFromSigned(signed, cagent)
        d.addCallback(lambda x, entry=self: entry)
        return d

    def updateIDRecord(self, signed, cause):
        """Update this entry with the signed IDRecord they just sent us.
        Return a Deferred that will fire (with this entry as an argument)
        when the update has completed."""
        # TODO: record 'cause' as the cause of the update
        d = self.idRecord.updateFromSigned(signed)
        d.addCallback(self._updated)
        return d

    def _updated(self, results):
        self.notifyWatchers()
        return self

    def updateOurPermission(self, encoded):
        """Accept an encoded permission slip (usually from a
        'yourPermission' message node) and update the .ourPermission
        attribute with its contents.

        You are responsible for calling .modified() after this.
        """
        if not encoded:
            self.ourPermission = None
        else:
            if not self.ourPermission:
                self.ourPermission = permission.PermissionSlip()
            self.ourPermission.setFromEncoded(encoded)

    def _setOurPermission(self, perm):
        # for testing
        assert implements(perm, IPermissionSlip)
        self.ourPermission = perm

    def setTheirPermission(self, perm):
        """You are responsible for calling .modified() after this."""
        assert implements(perm, IOwnerPermission)
        self.theirPermission = perm
    def revokeTheirPermission(self):
        """You are responsible for calling .modified() after this."""
        self.theirPermission = None

    def modified(self):
        # tell the agent that this recipient needs to be refreshed at some
        # point
        self.notifyWatchers()
        if not self.addressbook:
            return
        if not self.doUpdate:
            return
        owner = self.addressbook.owner
        agent = owner.parent
        why = "Must update recipient %s with their new settings" % \
              self.getName()
        d = agent.scheduleEvent(self.updateHoldoffTime, why, owner.ID)
        d.addCallback(agent.refreshRecipient, self.idRecord.ID)


    def getName(self):
        if self.petname != None:
            return self.petname
        if self.idRecord and self.idRecord.name != None:
            return '"%s"' % self.idRecord.name
        if self.idRecord:
            return self.idRecord.ID
        return "????"

    def getTransport(self, agent, message):
        """Obtain a Transport object which can be used to deliver an
        encapsulated message to this recipient.
        """
        t = transport.transportFromString(self.idRecord.transport, agent)
        return t

    def encodePermissionSlip(self):
        """Return a string with an encoded copy of .theirPermission, which
        can be given to the sender. Returns None if they do not have
        specific permission.
        """

        if self.theirPermission:
            return self.theirPermission.encode()
        return None


class IDRecord(object):
    """This record describes a recipient. Each recipient in the agent's
    addressbook has one. Each is uniquely identified by the hash of the
    public key.

    The owner of the agent has some OwnerEntry records, and these are
    published in various ways. When somebody looks up one of those published
    records, the data they retrieve is turned into an IDRecord.
    
    Serialized forms of these records are retrieved in the process of
    sending a message. This is the form in which they are stored in the
    addressbook and the 'me' records.
    """

    __implements__ = IIDRecord

    def __init__(self):
        self.ID = None     # hash of the public key
        self.pubkey = None # public key
        # identifying information (all optional)
        self.name = ""
        self.addresses = [] # email addresses which point to this record
        # permission record
        self.genericPermission = None # non-sender-specific permissions
        # transport record
        self.transportString = None
        # other stuff
        self.expires = None
        self.signed = None

    def updateGenericPermission(self, encoded):
        if not self.genericPermission:
            self.genericPermission = permission.PermissionSlip()
        self.genericPermission.setFromEncoded(encoded)

    def createFromSigned(self, signed, ca):
        """Take a signed/encoded form (the output of OwnerEntry.sign) and
        populate this record. Do not add the key to the crypto agent. Do not
        try to verify the consistency of the record (like it being signed by
        the public key it provides)."""
        assert signed
        d = ca.verify(signed)
        d.addCallback(self._createFromSigned2)
        return d

    def _createFromSigned2(self, results):
        sigID, encoded = results
        # sigID will probably be None, as this is a new key
        record = util.decode(encoded)
        self.ID = record['ID']
        self._update(record)
        return self

    def addFromSigned(self, signed, ca):
        """Take a signed/encoded form (the output of OwnerEntry.sign) and
        populate this record."""
        assert signed
        d = ca.verify(signed)
        d.addCallback(self._addFromSigned2, signed, ca)
        return d

    def _addFromSigned2(self, results, signed, ca):
        sigID, encoded = results
        # sigID will probably be None, as this is a new key
        record = util.decode(encoded)
        d = ca.addPubkey(record['pubkey'])
        d.addCallback(self._addFromSigned3, signed, ca, record)
        return d

    def _addFromSigned3(self, pubkey, signed, ca, record):
        # now that the key has been added to the keyring, verify it again
        d = ca.verify(signed)
        d.addCallback(self._addFromSigned4, pubkey, signed, record)
        return d

    def _addFromSigned4(self, results, pubkey, signed, record):
        sigID, dummy = results
        if pubkey.getID() != sigID:
            raise ValueError("IDRecord signed by the wrong key:" \
                             "key is %s, sig is %s" % (pubkey.getID(),
                                                       sigID))
        if record['ID'] != sigID:
            raise ValueError("IDRecord holds ID field with wrong key:" \
                             "ID field is %s, sig is %s" % (record['ID'],
                                                            sigID))
        self.ID = record['ID']
        self.pubkey = pubkey
        self.signed = signed # to send to others
        self._update(record)
        return self

    def updateFromSigned(self, signed):
        ca = self.pubkey.agent
        d = ca.verify(signed)
        d.addCallback(self._updateFromSigned2, signed)
        return d

    def _updateFromSigned2(self, results, signed):
        ca = self.pubkey.agent
        sigID, encoded = results
        if sigID != self.ID:
            raise ValueError("update not correctly signed")
        record = util.decode(encoded)
        if record['ID'] != self.ID:
            raise NotImplementedError("pubkey update")
        d = ca.addPubkey(record['pubkey'])
        d.addCallback(self.updateFromSigned3, signed, record)
        return d

    def updateFromSigned3(self, newpubkey, signed, record):
        if newpubkey.getID() != self.ID:
            # TODO: handle key changes
            raise NotImplementedError("pubkey update")
        self.signed = signed # to send to others
        self._update(record)
        return self

    def _update(self, d):
        self.name = d['name']
        self.addresses = d['addresses']
        self.expires = d['expires']
        slip = d.get('permission')
        if slip:
            self.updateGenericPermission(slip)
        transports = d['transports']
        self.transportString = None
        if transports:
            self.transportString = transports[0] # TODO

    def getTransport(self, agent, message):
        tport = transport.transportFromString(self.transportString, agent)
        return defer.succeed(tport)

class FakeNameServer(object):
    def __init__(self):
        self.names = {}
        self.IDs = {}
    def publish(self, name, record):
        d = record.sign()
        d.addCallback(self.publish2, name, record)
        d.addErrback(log.err)
    def publish2(self, signed, name, record):
        self.names[name] = signed
        self.IDs[record.ID] = signed

    def lookup(self, agent, name):
        r = IDRecord()
        return r.addFromSigned(self.names.get(name), agent.crypto)
    def lookupID(self, agent, ID):
        r = IDRecord()
        return r.addFromSigned(self.IDs.get(ID), agent.crypto)
    def describe(self):
        return "Fake"

class DiskNameServer(object):
    def __init__(self, args):
        self.mapdir = args

    def _getAllRecords(self, agent, records):
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        defs = []
        for f in os.listdir(mapdir):
            print "checking", f
            s = open(os.path.join(mapdir, f), "r").read()
            r = IDRecord()
            d = r.addFromSigned(s, agent.crypto)
            d.addCallback(self._gotARecord, records)
            d.addErrback(self._notGotARecord, f)
            defs.append(d)
        dl = defer.DeferredList(defs)
        return dl

    def _gotARecord(self, record, records):
        print " record was", record.name, record.ID
        records.append(record)
    def _notGotARecord(self, failure, f):
        print "DiskNameServer._getAllRecords failed on", f
        log.err(failure)
        return None

    def lookup(self, agent, address):
        """Take a name and return a (single) ID Record, using a server of
        some sort."""
        print "DiskNameServer.lookup(%s)" % address
        records = []
        d = self._getAllRecords(agent, records)
        d.addCallback(self._doLookup, records, address)
        return d
    def _doLookup(self, results, records, address):
        for r in records:
            if address in r.addresses:
                return r
        return None

    def lookupID(self, agent, ID):
        """Take an ID number and return a (single) ID Record, using a server
        of some sort."""
        records = []
        d = self._getAllRecords(agent, records)
        d.addCallback(self._doLookup, records, ID)
        return d
    def _doLookupID(self, results, records, ID):
        for r in records:
            if ID == r.ID:
                return r
        return None

    def publish(self, address, record):
        d = record.sign()
        d.addCallback(self.publish2, address, record)
        d.addErrback(log.err)
    def publish2(self, signed, address, record):
        mapdir = os.path.expanduser(self.mapdir)
        if not os.path.exists(mapdir):
            os.mkdir(mapdir)
        f = open(os.path.join(mapdir, record.ID), "w")
        f.write(signed)
        f.close()

    def describe(self):
        return "Disk:%s" % self.mapdir

class PBNameServer(object):
    def __init__(self, url):
        self.hostname, portnum = url.split(":")
        self.portnum = int(portnum)

    def lookup(self, agent, address):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._lookup2, "lookup", agent, address)
        return d

    def lookupID(self, agent, ID):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._lookup2, "lookupID", agent, ID)
        return d

    def _lookup2(self, server, method, agent, address):
        d = server.callRemote(method, address)
        d.addCallback(self._lookup3, agent)
        return d

    def _lookup3(self, s, agent):
        r = IDRecord()
        d = r.addFromSigned(s, agent.crypto)
        return d

    def publish(self, address, record):
        d = pbutil.connect(self.hostname, self.portnum)
        d.addCallback(self._publish2, address, record)
        return d
    def _publish2(self, server, address, record):
        d = record.sign()
        d.addCallback(self._publish3, server)
        return d
    def _publish3(self, signed, server):
        d = server.callRemote("publish", signed)
        return d

    def describe(self):
        return "PB:%s:%d" % (self.hostname, self.portnum)    

def getLookupAgentFromString(args):
    scheme, rest = args.split(":", 1)
    fac = {"Fake": FakeNameServer,
           "Disk": DiskNameServer,
           "PB": PBNameServer,
           }
    return fac[scheme](rest)
