#! /usr/bin/python

import sys, re, os
from twisted.trial import unittest
from twisted.python.components import implements
from twisted.internet import reactor, defer
dr = unittest.deferredResult

import petmail.crypto
petmail.crypto.EncryptionAgent = petmail.crypto.FakeEncryptionAgent

from petmail import addressbook, permission, agent, interfaces, message, \
     owner, crypto, transport, util, processors

class FakeAgent(agent.Agent):
    def acceptRFC822(self, rm, why):
        self.inbox.append((rm,why))
    def scheduleEvent(self, *args):
        return defer.Deferred() # will never fire

class NonLiveTest(unittest.TestCase):

    def build(self):
        server = addressbook.FakeNameServer()

        self.crypto = crypto.EncryptionAgent()

        alice_agent = FakeAgent('aliceAgent')
        alice_agent.registerAddressLookupAgent(server)
        alice_agent.populateProcessors()
        alice_agent.crypto = self.crypto

        alice = owner.OwnerEntry()
        alice.setKey(dr(self.crypto.generateKey("alice"), 60))
        alice_agent.addOwner(alice)

        alice_agent.addressbook = addressbook.AddressBook(alice)
        
        alice.setName("Alice")
        alice.addAddress("alice@foo.com")
        alice.setGenericPermission(permission.OwnerPermission())
        alice.setInlet(transport.DummyInlet("alice"))
        alice.publish()

        alice_agent.registerAddressLookupAgent(server)

        bob_agent = FakeAgent("bobAgent")
        bob_agent.registerAddressLookupAgent(server)
        bob_agent.populateProcessors()
        bob_agent.crypto = self.crypto

        bob = owner.OwnerEntry()
        bob.setKey(dr(self.crypto.generateKey("bob"), 60))
        bob_agent.addOwner(bob)

        bob_agent.addressbook = addressbook.AddressBook(bob)

        bob.setName("Bob")
        bob.addAddress("bob@foo.com")
        bob.setGenericPermission(permission.OwnerPermission())
        bob.setInlet(transport.DummyInlet("bob"))
        bob.publish()
        bob_agent.registerAddressLookupAgent(server)

        return (alice_agent, alice), (bob_agent, bob), server

    def test1(self):
        (alice_agent, alice), (bob_agent, bob), server = self.build()
        d = alice_agent.retrieveAddress("bob@foo.com")
        bob_record = unittest.deferredResult(d)
        self.failUnless(isinstance(bob_record, addressbook.IDRecord))

        body = \
"""From: alice@foo.com
To: bob@foo.com
Subject: Yar

Testing
 -Alice
"""
        bob_inbox = []
        bob_agent.inbox = bob_inbox
        #alice_agent.sendMessage0(alice, bob_record, body)
        m = message.Message(alice_agent)
        m.setSenderEntry(alice)
        m.setRecipientRecord(bob_record)
        #m.setRecipientEntry(None)
        m.recipientEntry = None; m.provide('recipientEntry')
        m.addRFC822(body)
        sent = []
        def deliver(msg, s=sent):
            s.append(msg)
            #print "deliver", msg
            #print "transport", msg.transport
        #alice_agent.debugProcessing = True

        d = m.startProcessing()
        sent = unittest.deferredResult(d)
        self.failUnlessIdentical(sent, m)
        self.failUnless(sent.getNode("messageID"))
        self.failUnless(sent.getNode("rfc822"))

    def test2(self):
        (alice_agent, alice), (bob_agent, bob), server = self.build()
        d = alice_agent.retrieveAddress("bob@foo.com")
        bob_record = unittest.deferredResult(d)
        self.failUnless(isinstance(bob_record, addressbook.IDRecord))

        body = \
"""From: alice@foo.com
To: bob@foo.com
Subject: Yar

Testing
 -Alice
"""
        bob_inbox = []
        bob_agent.inbox = bob_inbox
        m = message.Message(alice_agent)
        m.setSenderEntry(alice)
        m.setRecipientRecord(bob_record)
        #m.setRecipientEntry(None)
        m.recipientEntry = None; m.provide('recipientEntry')
        m.addRFC822(body)

        #alice_agent.debugProcessing = True
        #bob_agent.debugProcessing = True

        d = m.send()
        dr(d)

        self.failUnlessEqual(len(bob_inbox), 1)
        inmsg,why = bob_inbox[0]

        self.failUnless(implements(inmsg, interfaces.IReceivedMessage))
        self.failUnlessEqual(inmsg.headers['rfc822']['body'], body)
        self.failUnlessEqual(why, "(generic) no ticket required")



body = "A message body goes here"

class SingleAgent(unittest.TestCase):

    def setUp(self):
        self.oldUpdateHoldoffTime = None

    def tearDown(self):
        self.cleanAgentTimer()
        if self.oldUpdateHoldoffTime != None:
            old = self.oldUpdateHoldoffTime
            addressbook.AddressBookEntry.updateHoldoffTime = old

    def cleanAgentTimer(self):
        if self.agent.scheduledEventTimer:
            self.agent.scheduledEventTimer.cancel()
            self.agent.scheduledEventTimer = None

    def buildAgent(self):
        server = addressbook.FakeNameServer()
        self.agent = FakeAgent('singleAgent')
        self.agent.populateProcessors()
        self.agent.crypto = crypto.EncryptionAgent()
        self.agent.registerAddressLookupAgent(server)
        self.agent.inbox = []

    def buildIdentity(self, name, address):
        assert " " not in name  # also used as a directory for generateKey

        us = dr(self.agent.createIdentity())
        us.setName(name)
        us.addAddress(address)
        us.setGenericPermission(permission.OwnerPermission())
        us.setInlet(transport.DummyInlet(name))
        us.grantReturnPermission = False
        self.agent.addOwner(us)
        us.publish()
        return us

    def build(self):
        self.buildAgent()
        alice = self.buildIdentity("alice", "alice@example.com")
        bob = self.buildIdentity("bob", "bob@example.com")
        return (alice, bob)

    def getIDRecord(self, dest):
        address = dest.addresses[0]
        idrecord = dr(self.agent.retrieveAddress(address))
        return idrecord

    def addToAddressBook(self, owner, dest):
        signed = dr(dest.sign())
        entry = addressbook.AddressBookEntry()
        dr(entry.createIDRecord(signed, self.agent.crypto, "no reason"))
        self.agent.addressbooks[owner].addEntry(entry)
        return entry

    def buildMessage1(self, alice, bob):
        to_bob = self.addToAddressBook(alice, bob)
        
        m = message.Message(self.agent)
        m.setSenderEntry(alice)
        m.setRecipientEntry(to_bob)
        m.addRFC822(body)

        return m

    def test1(self):
        alice, bob = self.build()
        m = self.buildMessage1(alice, bob)

        dr(self.agent.sendMessage(m))

        # at this point the message has been fully processed (both by the
        # sender and, due to the DummyTransport, also by the receiver).

        self.failUnlessEqual(len(self.agent.inbox), 1)
        m1 = self.agent.inbox[0][0]
        self.failUnless("checkedRFC822" in m1.provides)
        self.failUnless(m1.allowRFC822)
        self.failUnless(m1.headers['rfc822'])
        self.failUnless(m1.headers['rfc822']['body'] == body)

        # bob doesn't know who the message is from
        self.failIf(m1.senderEntry)

        # alice should not have given permission to bob
        to_bob = self.agent.addressbooks[alice].lookupID(bob.ID)
        self.failUnless(to_bob)
        self.failIf(to_bob.theirPermission)

        # make sure bob can add the entry
        self.failUnless(m1.canAddSender())
        dr(m1.addSender())
        self.failUnless(m1.senderEntry)

    def test2(self):
        # verify that alice gives permission to Bob
        alice, bob = self.build()
        alice.grantReturnPermission = True
        m = self.buildMessage1(alice, bob)

        dr(self.agent.sendMessage(m))

        to_bob = self.agent.addressbooks[alice].lookupID(bob.ID)
        self.failUnless(to_bob)
        self.failUnless(to_bob.theirPermission)

        
    def test3(self):
        raise unittest.SkipTest("this functionality was changed")
        # verify that bob gives inbound permission to alice
        alice, bob = self.build()

        m = self.buildMessage1(alice, bob)

        dr(self.agent.sendMessage(m))

        back_to_alice = self.agent.addressbooks[bob].lookupID(alice.ID)
        self.failUnless(back_to_alice)
        self.failIf(back_to_alice.theirPermission)

        self.failUnless(len(self.agent.scheduledEvents), 1)
        ev = self.agent.scheduledEvents[0][1]
        self.failUnless("granting them permission to send more" in ev.why())
        #print "will fire in", self.agent.scheduledEvents[0][0] - util.now()
        
        #print "waiting for inbound permission holdoff"

        # now wait an additional three seconds for the holdoff-timer: 1 for
        # the timer, 1 for the "wake up a bit late" in checkScheduledEvents,
        # 1 for margin
        then = util.now() + 3
        while util.now() < then:
            reactor.iterate(0.1)
            
        self.failUnless(back_to_alice.theirPermission)

        # and there should be a scheduled event to update them
        self.failUnlessEqual(len(self.agent.scheduledEvents), 1)
        ev = self.agent.scheduledEvents[0][1]
        self.failUnless("Must update recipient" in ev.why())

    def test4(self):
        raise unittest.SkipTest("this functionality was changed")
        # verify that alice accepts that inbound permission
        alice, bob = self.build()
        bob.grantInboundPermission = True
        for p in self.agent.inboundProcessors:
            if isinstance(p, processors.GrantInboundPermission):
                p.permissionGrantHoldoffTime = 1
        # accelerate the update process. Must set this on the class because
        # a new instance will be created for the inbound message
        old = addressbook.AddressBookEntry.updateHoldoffTime
        self.oldUpdateHoldoffTime = old
        addressbook.AddressBookEntry.updateHoldoffTime = 1
        
        m = self.buildMessage1(alice, bob)

        dr(self.agent.sendMessage(m))

        back_to_alice = self.agent.addressbooks[bob].lookupID(alice.ID)
        self.failUnless(back_to_alice)
        self.failIf(back_to_alice.theirPermission)

        self.failUnless(len(self.agent.scheduledEvents), 1)
        ev = self.agent.scheduledEvents[0][1]
        self.failUnless("granting them permission to send more" in ev.why())
        #print "will fire in", self.agent.scheduledEvents[0][0] - util.now()
        
        #print "waiting for inbound permission holdoff"

        # now wait an additional three seconds for the holdoff-timer: 1 for
        # the timer, 1 for the "wake up a bit late" in checkScheduledEvents,
        # 1 for margin
        then = util.now() + 3
        while util.now() < then:
            if back_to_alice.theirPermission:
                break
            reactor.iterate(0.1)
        self.failUnless(back_to_alice.theirPermission)

        to_bob = self.agent.addressbooks[alice].lookupID(bob.ID)
        self.failIf(to_bob.ourPermission)
        # now wait three seconds for the return message to propogate
        then = util.now() + 3
        while util.now() < then:
            if to_bob.ourPermission:
                break
            reactor.iterate(0.1)
        self.failUnless(to_bob.ourPermission)
